#!/usr/bin/env python3

from __future__ import print_function
import keras
from keras.datasets import cifar10
from keras.preprocessing.image import ImageDataGenerator
from keras.models import Sequential
from keras.layers import Dense, Dropout, Activation, Flatten, Lambda
from keras.layers import Conv2D, MaxPooling2D
import os
import numpy as np
from keras import backend as K
from scipy.sparse.linalg.interface import _AdjointMatrixOperator
from os.path import join, isdir
import common


import keras.losses
import keras.metrics

import scipy.misc
import scipy.ndimage
import skimage.transform
from skimage.morphology import watershed
import argparse
import tifffile


smooth = 1.


def dice_coef(y_true, y_pred):
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return (2. * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)


def dice_coef_loss(y_true, y_pred):
    return -dice_coef(y_true, y_pred)


def masked_mean_squared_error(y_true, y_pred):
    mask = K.cast(Lambda(lambda x: x > -10, output_shape=(y_pred.shape[1], y_pred.shape[2], y_pred.shape[3]))(y_true),
                  'float32')
    diff = (y_true - y_pred) * mask
    squared_difference = K.sum(diff * diff)  # K.sum(((y_true - y_pred) * (y_true - y_pred)) * mask)
    squared_difference /= K.sum(mask)
    return squared_difference


def masked_mean_abs_error(y_true, y_pred):
    mask = K.cast(Lambda(lambda x: x > -10, output_shape=(y_pred.shape[1], y_pred.shape[2], y_pred.shape[3]))(y_true),
                  'float32')
    diff = (y_true - y_pred) * mask
    squared_difference = K.sum(K.abs(diff))  # K.sum(((y_true - y_pred) * (y_true - y_pred)) * mask)
    squared_difference /= K.sum(mask)
    return squared_difference


keras.metrics.dice_coef = dice_coef
keras.losses.dice_coef_loss = dice_coef_loss

keras.losses.masked_mean_squared_error = masked_mean_squared_error
keras.losses.masked_mean_abs_error = masked_mean_abs_error

K.set_image_data_format('channels_last')  # TF dimension ordering in this code




def get_base_weight(shape, unreliable_d=None):
    if (unreliable_d is None):
        unreliable_d = int(shape[0] / 5)
    base_weights = np.zeros(shape, dtype=bool)
    base_weights[1:-1, 1:-1] = True
    base_weights = scipy.ndimage.distance_transform_edt(base_weights)
    base_weights -= unreliable_d
    base_weights /= unreliable_d
    base_weights[base_weights < 0.000001] = 0.000001
    base_weights[base_weights > 1] = 1
    # scipy.misc.imsave('test/base_weights.png', base_weights)
    return base_weights

def apply_patches(model, config, all_pred_np, sum_np, patchs_to_process, imgs, base_weights, data_generator_module):
    imgs = np.array(imgs)
    imgs = data_generator_module.process_X(imgs)

    pred_np = model.predict(imgs, verbose=1)

    for i in range(len(patchs_to_process)):
        ex_pred_np = all_pred_np[patchs_to_process[i][0]:patchs_to_process[i][1], patchs_to_process[i][2]:patchs_to_process[i][3]]
        p_np = pred_np[i, :, :]
        if (hasattr(config, 'output_postprocessing')):
            p_np = config.output_postprocessing(p_np)
        ex_pred_np += p_np * base_weights / sum_np[patchs_to_process[i][0]:patchs_to_process[i][1], patchs_to_process[i][2]:patchs_to_process[i][3]]

def predict_data(model, config, X, batch_size=32):
    data_generator_module = __import__('data_generators.' + config.generator, fromlist=[''])
    image_rows = config.input_shape[0]
    image_cols = config.input_shape[1]
    nb_dim_in = config.input_shape[2]
    nb_dim_out = config.output_shape[2]
    patch_shape = (image_rows, image_cols)

    X = config.image_preprocessing(X)



    #X = X[:,:,np.newaxis]

    X_pad = [(0, max(0, patch_shape[0] - X.shape[0])), (0, max(0, patch_shape[1] - X.shape[1]))]
    if (len(X.shape) == 3):
        X_pad.append((0, 0))
    X = np.lib.pad(X, X_pad,
                   'constant', constant_values=(0))


    patches = common.generate_patches(X.shape[:2], (image_rows, image_cols))

    base_weights = get_base_weight((image_rows, image_cols, nb_dim_out))

    all_pred_np = np.zeros((X.shape[0], X.shape[1], nb_dim_out), dtype='float32')
    sum_np = np.zeros((X.shape[0], X.shape[1], nb_dim_out), dtype='float32')
    for i in range(len(patches)):
        sum_np[patches[i][0]:patches[i][1], patches[i][2]:patches[i][3]] += base_weights


    

    imgs = []
    patchs_to_process = []
    for patch in patches:
        ex_X = config.process_patch(X[patch[0]:patch[1], patch[2]:patch[3]], False)
        imgs.append(ex_X)
        patchs_to_process.append(patch)
        if (len(patchs_to_process) >= batch_size):
            apply_patches(model, config, all_pred_np, sum_np, patchs_to_process, imgs, base_weights, data_generator_module)
            imgs = []
            patchs_to_process = []

    if (len(patchs_to_process) > 0):
        apply_patches(model, config, all_pred_np, sum_np, patchs_to_process, imgs, base_weights, data_generator_module)

    #all_pred_np = np.clip(all_pred_np, 0, 1)
    print (np.min(all_pred_np), np.max(all_pred_np))
    return X, all_pred_np


def predict(model, config_filepath, path):
    config = __import__(config_filepath, fromlist=[''])
    
    if (path.endswith('.tif')):
        X = tifffile.imread(path)
    elif (path.endswith('.npz')):
        X = np.load(path)['X']
    else:
        X = scipy.misc.imread(path)
    
    
    return predict_data(model, config, X)


add_image = False
# Kitti2015_000002_10
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Train on patches')
    parser.add_argument('config', type=str, help='Configuration')
    parser.add_argument('input', type=str, help='Image')
    parser.add_argument('output', type=str, help='Output')
    parser.add_argument('--model', type=str, help='Model')
    parser.add_argument('--debug', type=bool, default=False, help='Save debug images in tmp folder')

    args = parser.parse_args()

    config = __import__(args.config, fromlist=[''])

    model_predict = args.model
    if (model_predict is None):
        model_predict = config.model_predict

    model = keras.models.load_model(model_predict)

    im_np, pred_np = predict(model, args.config, args.input)

    np.savez_compressed(args.output, im_np=im_np, pred_np=pred_np)

    if (args.debug):
        tmp_folder = 'tmp'

        if (not isdir(tmp_folder)):
            os.makedirs(tmp_folder)

        scipy.misc.imsave(join(tmp_folder, 'im.png'), im_np)#im_np[:,:,0:3]
        #for i in range(im_np.shape[2]):
        #    scipy.misc.imsave(join(tmp_folder, 'im_'+str(i)+'.png'), im_np[:, :, i])
            
        '''
        y_np = np.load(args.input[:-5] + 'y.npz')['y']
        print (y_np.shape)
        scipy.misc.imsave(join(tmp_folder, 'y.png'), y_np)
        y_np_new = []
        classes =  [2, 5, 6, 9, 17, 65]
        for c in classes:
            y_np_new.append(y_np == c)
           
        y_np = np.transpose(np.asarray(y_np_new).astype('float'), (1, 2, 0))
        for i in range(y_np.shape[2]):
            scipy.misc.imsave(join(tmp_folder, 'y_'+str(i)+'.png'), y_np[:,:,i])
        '''
        scipy.misc.imsave(join(tmp_folder, 'pred.png'), np.argmax(pred_np, axis=2))
        for i in range(pred_np.shape[2]):
            scipy.misc.imsave(join(tmp_folder, 'pred_'+str(i)+'.png'), (pred_np[:, :, i] > 0).astype(float))
            tifffile.imsave(join(tmp_folder, 'pred_'+str(i)+'.tif'), pred_np[:, :, i] * 255)
